/*
  Objective:
	To test the assertion as required by FIPS 151-2.
	"If a read() on a terminal device is interrupted by a signal after
	it has successfully read some data, it shall return the number of
	bytes the system has read."

	Two possibilities exist:
	  1) Test with VMIN > 0 and VTIME = 0 for the following two possibilities;
			a) Call to read() when interrupted while blocked, returns the
			   data that has been read.

			b) Call to read() when interrupted while blocked, returns -1
			   and sets errno to [EINTR].

				Once the read() is blocking (waiting for input) a signal will
				abort the read() and return "-1" with errno set to [EINTR].
				A following read() will get all data queued up and no data
				will be lost.

	  2) Test with VMIN = 0 and VTIME = 0:
			The problem with attempting to interrupt the read() is a timing
			problem.  Was the read() interrupted, or did only that much data
			exist in the queue at the time of the read().  I don't believe a
			programming solution exists for resolving this question with what
			POSIX.1 provides.

	Test strategy implemented: 1a and 1b.
*/


#include	<pcts.h>

#include	<Zunistd.h>
#include	<sys/Ztimes.h>
#include	<Zfcntl.h>
#include	<Ztermios.h>

#include	<sv_env.h>
#include	<sv_errno.h>
#include	<pcts_timer.h>


#define	START_VALUE	0
#define	ALL_RIGHT	2
#define	TOO_LATE	3

#define	FILLER		'Z'
#define	DATA		'D'
#define	L_DATA		'S'

#define	TIME_TO_READ	4

static	char	*buf_in;
static	size_t	max_input, nb_write;
static	int		ct_sleep(clock_t);

int	indx;
void sigs_int(int i)
{
	if(buf_in[nb_write -1] == FILLER)
		indx = ALL_RIGHT;
	else
		indx = TOO_LATE;
}

int main(void)
{
	char	path[_POSIX_PATH_MAX +1];
	pid_t	pid;
	char	*buf_out;
	int		s_errno;
	int		fd;
	int		tty_fildes, loop_fildes;
	struct	termios tty_termios, *tty_termios_p = &tty_termios,
					loop_termios, *loop_termios_p = &loop_termios;
	clock_t	t_read;
	clock_t	clk_tcks;
	int		r_cnt1, r_cnt;
	int		i;

	setup(E_READ);
	funct(E_READ);

	assertion(12, NOT_APPLICABLE);
	assertion(13, NOT_APPLICABLE);

	assertion(14, EXTENDED);
	assertion(15, EXTENDED);
	assertion(16, NOT_APPLICABLE);
	assertion(17, EXTENDED);

	exit_on(ERROR);

	assertion(11, BASE);
	start
		GTI_NOT_PROVIDED_EXIT

		prep("Establish path to temp file for passing parameters.");
		start
			Zstrcpy(path, temp_path());
			Zstrcat(path, "int_read");
		finis

		max_input = g_max_input(PORT_TTY);
		if(max_input > (size_t)((cc_t)-1)) /* Limit testing to size of VMIN */
			max_input = (size_t)((cc_t)-1);

		prep("Establish PORT_TTY as the control terminal and PORT_LOOP.");
		prep("Set speed for ports to lowest speed supported (B%d).",
						(int)baud_value(baud_min()) );
		start
			termios_prep(&tty_fildes, tty_termios_p, &loop_fildes,
							loop_termios_p);

				/*	Compute number of bytes to write in TIME_TO_READ seconds
					and the time-to-read in clock-ticks.
					If the TIME_TO_READ results in more than max_input bytes,
					then only read max_input bytes and recompute time-to-read.
				*/
			clk_tcks = Zsysconf(_SC_CLK_TCK);
			t_read = TIME_TO_READ * clk_tcks;
			nb_write = (size_t) (baud_value(baud_min())/(sizeof(char)*8 +1)) *
												TIME_TO_READ;
			if(nb_write > max_input) {
				nb_write = max_input;
				t_read = (size_t)(( (max_input/baud_value(baud_min())) /
										(sizeof(char)*8 +1) )*clk_tcks);
				if(t_read == 1)
					error("Cannot test at speed of %d and {MAX_INPUT} of %ld.",
								(int)baud_value(baud_min()),
								(unsigned long)max_input);
				if(t_read < 4)
					warning("Timing tight, speed of %d and {MAX_INPUT} of %ld.",
								(int)baud_value(baud_min()),
								(unsigned long)max_input);
			}
			buf_in = (char *) Zmalloc(nb_write +1);
			buf_out = (char *) Zmalloc(nb_write +1);

			tty_termios_p->c_lflag &= ~(ICANON | ECHO);
			tty_termios_p->c_cc[VMIN] = (cc_t) nb_write;
			tty_termios_p->c_cc[VTIME] = 0;
			Zcfsetospeed(loop_termios_p, baud_min());
			Zcfsetispeed(loop_termios_p, baud_min());
			Zcfsetospeed(tty_termios_p, baud_min());
			Zcfsetispeed(tty_termios_p, baud_min());
			Ztcsetattr(tty_fildes, TCSANOW, tty_termios_p);
			Ztcsetattr(loop_fildes, TCSANOW, loop_termios_p);
		finis

		prep("Set VTIME and VMIN to wait till %lu bytes",
					(unsigned long) nb_write);
		prep("are read or an interrupt occurs.");
		prep("Time to transmit should be %d clock-ticks (%d clock-ticks/sec).",
						(int)t_read, (int)clk_tcks);
		start
		finis


		prep("initialize input and output data buffers.");
		start
			memset(buf_in, FILLER, nb_write);
			memset(buf_out, DATA, nb_write);
			buf_out[nb_write -1] = L_DATA;
		finis

		ENABLE_PSYNC

		prep("Fork a child.");
		start
			sv_flush();
			pid = Zfork();
		finis

		if(pid == (pid_t)0) {
			prep("Issue read() while expecting an interrupt.");
			start
				indx = START_VALUE;
				SIGNAL(SIGINT, sigs_int);

				SEND_PSYNC

				errno = 0;
							/*	read() should not complete until nb_write
								bytes are read or an interrupt occurs.
							*/
				r_cnt1 = read(tty_fildes, buf_in, nb_write);
				s_errno = errno;
			finis

			prep("Save parameters for testing in parent.");
			start
				fd = Zcreat(path, PROT_ALL);
				Zwrite(fd, (char *)&r_cnt1, sizeof(int));
				Zwrite(fd, (char *)&s_errno, sizeof(int));
				Zwrite(fd, (char *)&indx, sizeof(int));
				if(r_cnt1 > 0)
					Zwrite(fd, buf_in, r_cnt1);
				Zclose(fd);
			finis

			done();
		}

							/* Wait for parent to start the read(). */
		SET_TIMEOUT
			WAIT_PSYNC
		CLEAR_ALARM

		prep("Issue call to transmit %lu bytes to PORT_LOOP.",
										(unsigned long) nb_write);
		prep("Delay then send SIGINT.");
		start
			expecting(nb_write);
			Zwrite(loop_fildes, buf_out, nb_write);
					/*	Delay time needed to transmit half the data.  */
			ct_sleep(t_read/2);

			if(kill(pid, SIGINT) == -1) {
				error("Call to kill(), obtained errno of %s\n",
													err_lookup(errno));
				done();
			}
			Ztcdrain(loop_fildes);		/* Transmit rest of data. */
		finis

		DISABLE_PSYNC

		WAIT_FOR_PID(pid);

		prep("Get child parameters.");
		start
			fd = Zopen(path, O_RDONLY);
			Zread(fd, (char *)&r_cnt1, sizeof(int));
			Zread(fd, (char *)&s_errno, sizeof(int));
			Zread(fd, (char *)&indx, sizeof(int));
			if(r_cnt1 > 0)
				Zread(fd, buf_in, r_cnt1);
			Zclose(fd);
			Zunlink(path);
		finis

		if(PCTS_GTI_BUFFERS_OUTPUT) {
			prep("Ensure read() was interrupted.");
			start
				if(indx == START_VALUE)
					error("Call to read() was not interrupted.");

						/* UNRESOLVED:  Did not properly interrupt ??? */
				if(indx == TOO_LATE)
					error("Interrupted too late.");
			finis
		}

		prep("Check if unexpected error occurred.");
		start
			if(r_cnt1 == -1)
				if(s_errno != EINTR)
					error("%s\n%s\n\t\t%s\n",
						"Call to read() returned unexpected results.",
						"Call to  read() returned -1 with errno set to:",
						err_lookup(s_errno));
				else
					message("Call to read() returned -1 with errno set EINTR.");
		finis

		if(r_cnt1 == nb_write) {
				/*	The read() was interrupted but then continued and
					successfully read all data required to be read.

					Or

					The implementation does not buffer the output,
					and the interrupt is not processed prior to read()
					completion.
				*/
			testing("For FIPS 151-2 requirement Item \"m\".");
			testing("Call of read() when interrupted returned");
			testing("proper count of data read.");
			start
				for(i=0; i<nb_write -1; i++)
					if(buf_in[i] != DATA) {
						error("Incorrect count returned.");
						done();
					}

				if(buf_in[nb_write -1] != L_DATA) {
					error("Incorrect count returned.");
					done();
				}
			finis

			done();			/* Successfully completed FIPS */
		}

		testing("For FIPS 151-2 requirement Item \"m\".");
		if(r_cnt1 == -1) {
			testing("When a call to read() of a terminal device interrupted");
			testing("by a signal returns -1 and sets errno to [EINTR],");
			testing("a subsequent read() will return the %d bytes initially",
							(int)nb_write);
			testing("written to the device.");
		}
		else {
			testing("When a call to read() of a terminal device interrupted");
			testing("by a signal returns %d bytes, a subsequent read() returns",
						(int)r_cnt1);
			testing("the remaining %d bytes initially written to the device.",
							(int)(nb_write -r_cnt1));
		}
		start
    		size_t bytes_read = 0;
							/* Enable a read() for data available in queue. */
			tty_termios_p->c_cc[VMIN] = (cc_t) 0;
			Ztcsetattr(tty_fildes, TCSANOW, loop_termios_p);

			SET_TIMEOUT
				r_cnt = nb_write -r_cnt1;
				if(r_cnt1 == -1) {
					r_cnt = nb_write;
					r_cnt1 = 0;
				}
				while(bytes_read < r_cnt) {
					bytes_read += Zread(tty_fildes, buf_in +r_cnt1 + bytes_read,
										(size_t)r_cnt);
					dt_secs(TIME_ACQUIRE_XMITTED_DATA);
				}
			CLEAR_ALARM

			for(i=0; i<nb_write -1; i++)
				if(buf_in[i] != DATA)
					error("Incorrect data received: buf_in[%d] contains %c",
									i, buf_in[i]);
			if(buf_in[nb_write -1] != L_DATA)
				error("Incorrect data received: buf_in[nb_write -1] contains %c"
									, buf_in[nb_write -1]);
		finis
	finis

	done();
}


		/*
			Clock ticks sleep, ct_sleep(clock_ticks_to_sleep)
			Delay elapsed time in CLK_TCKths of a second.
			Accuracy is limited to +0, -1 CLK_TCKths.
		*/
static	int ct_sleep(clock_t delay)
{
	struct	tms	tmsbuf; 
	clock_t	r_val;

	if(!delay)
		return(0);
	sleep(0);				/*	Place this process at end-of-queue for
								those proccesses waiting to run.

								Is this TRUE for all/any POSIX.1 systems???
							*/
	r_val = times(&tmsbuf);
	while(r_val +delay > times(&tmsbuf))
		;
	return(0);
}
